#include <TinyGPS++.h>
#include <SoftwareSerial.h>
#include <EEPROM.h>

// ====================================== GPS Values ===========================================

static const int TX_Pin = 4, RX_Pin = 3;  // these are the pins which the rx and tx of the gps are connected to. the tx pin of the gps is the most imprtant one and that pin number is used for the TX_Pin value (ie: tx to digital 4 means TX_Pin = 4)
static const uint32_t GPSBaud = 9600;     //this is the baud rate of the gps, this should apply to your gps but you might need to check the data sheet.

// ====================================== Moving Averge Values ===========================================

const int numReadings = 3;        // Number of readings for the running average 3 is a good balance between responsiveness and accuracy
float readings[numReadings];      
int readIndex = 0;               
float total = 0, MovingAverage = 0;                  

// ====================================== Speed Values ===========================================

float speed, speed2, maxSpeed, aveSpeed, sumSpeed = 0;

volatile int speedCount = 0;

// ====================================== Switch ===========================================

const int inputPin = 9;  // Pin connected to one side of the swicth
const int outputPin = 7;  // Pin connected to the other side of the switch

bool done, programingMode = false;

// ====================================== Setting up the gps serial ===========================================

TinyGPSPlus gps;                  //Creates a gps device for the TinyGPS+ library

SoftwareSerial ss(TX_Pin, RX_Pin);  //Connects the serial of the gps to get data

// ====================================== Setting up the code ===========================================

void setup() {
  Serial.begin(9600);           // Initialize serial communication with the computer if it is connected
  ss.begin(GPSBaud);            // Initialize serial communication with the gps if it is connected

  for (int i = 0; i < numReadings; i++) { // Initializes all moving average values to zero
    readings[i] = 0;
  }

  pinMode(inputPin, INPUT_PULLUP);    //sets up the switch pins ready to collect data
  pinMode(outputPin, OUTPUT);
  digitalWrite(outputPin, LOW); // make one pin of the switch high

  pinMode(LED_BUILTIN, OUTPUT);
  digitalWrite(LED_BUILTIN, HIGH);

  delay(1500);

  digitalWrite(LED_BUILTIN, LOW);

  delay(500); // allows 2 seconds before you cannot access programming mode

  if ((digitalRead(inputPin)) == LOW){ //activates prgramming mode
    programingMode = true;
    done = true;
  }
}

void loop() {

  if (((digitalRead(inputPin)) == HIGH) && (done == false)) { //if the switch has not be activated and the code is not finished it will gather data

    digitalWrite(LED_BUILTIN, LOW);

    // ====================================== Gathers Data ===========================================

    while (ss.available() > 0){       // if the gps is available it eneters a loop to collect data
      gps.encode(ss.read());          //Reads data
      if (gps.location.isUpdated()){  //If the data has been updated it proceeds to the next stage
        speed = gps.speed.kmph();     //Gets and calculates the speed value from the gps. Instead of .kmph() it could be .knots() or .mph() or .mps() depending on what units you want
      }
    }

    // ====================================== Begins Calculations ===========================================

    if (speed != speed2){     // if the speed value from the gps has changed 
      speed2 = speed;         // sets current speed as previous speed

      digitalWrite(LED_BUILTIN, HIGH);

      // ====================================== Moving Average Calcualtion ===========================================

      total = total - readings[readIndex];     // Subtract the last reading
      readings[readIndex] = speed;             // Sets the value to be averaged as the current speed value
      total = total + readings[readIndex];     // Add the reading to the total
      readIndex = readIndex + 1;               // Advance to the next position

      if (readIndex >= numReadings) {          // Wrap around when reaching the end of the array
        readIndex = 0;
      }

      MovingAverage = total / numReadings;     // Calculate the Moving average

      // ====================================== Other speed calculations ===========================================

      if (MovingAverage >= maxSpeed){       // if the Moving Average speed is larger than the maximum speed found previously the maximum speed is updated
        maxSpeed = MovingAverage;
      }

      sumSpeed = sumSpeed + MovingAverage;  // Calulates the total average using the moving average speed.
      speedCount = speedCount + 1;
      aveSpeed = (sumSpeed/speedCount);
    }

  } else if (done == false) { // if the switch is clicked and the code is not finished

  digitalWrite(LED_BUILTIN, HIGH);

  delay(500);

  digitalWrite(LED_BUILTIN, LOW);

  // ====================================== EEPROM ===========================================

    for (int i = 0 ; i < EEPROM.length() ; i++) { //clears eeprom memory
      EEPROM.write(i, 0);
    }

    EEPROM.put(0, maxSpeed); //writes the float values to the EEPROM memory
    EEPROM.put(5, aveSpeed);

    done = true; // code is now finished

    digitalWrite(LED_BUILTIN, HIGH);

  } else if (programingMode == true) { //if the switch was depressed when powering on

    digitalWrite(LED_BUILTIN, HIGH);

  // ====================================== Waiting for Computer signal ===========================================
  while (Serial.available() <= 0){
    Serial.println("Type Aything to Receive Data");
    delay(3500);
  }

  // ====================================== Recieveing stored data ===========================================

    Serial.println("");
    Serial.println("Programing Mode");
    Serial.println("");

    delay(1000);

    Serial.println("Maximum Speed Reached: ");
    EEPROM.get(0, maxSpeed); //gathers data from the stored memory
    Serial.print(maxSpeed); //prints the data
    Serial.println(" km/h");

    Serial.println("Average Speed: ");
    EEPROM.get(5, aveSpeed); //gathers data from the stored memory
    Serial.print(aveSpeed); //gathers data from the stored memory
    Serial.println(" km/h");

    programingMode = false; //Code is finished and programming mode is off
    done = true;
  }
}